#include <iostream>
#include <string>
#include <vector>
#include <sstream>
#include <getopt.h>
#include <cstdio>

using namespace std;

#include "tgvcnv.h"
#include "sys.h"
#include "tgv.h"
#include "debug.h"


struct conf_t
{
  bool same_con,decode;
  string bit_rate, filter;
  string audio_parm, video_parm;
  float frame_rate;
} conf;



//show usage
static
void usage(char *name)
{
    cout << "TGV converter " __TGV_VERSION__ ". (c)2017, psb^hlw.\n\nUsage: " <<
        name << " [options] input_file output_file\n\n"

        "Options:\n"
        "  -h           show this help\n"
        "  -f \"filt\"    specify ffmpeg filter instead of default one\n"
        "  -r framerate specify frame rate (12.25)\n"
        "  -b bitrate   specify sound bitrate (64k, 128k, 0 for copy as is)\n"
        "  -A \"...\"     additional ffmpeg parameter for audio processing\n"
        "  -V \"...\"     additional ffmpeg parameter for video processing\n"
        "  -d           show debug messages\n"
        "  -c           run ffmpeg in the same console\n"
        "  -t           convert TGV file to AVI\n"
        ;
}



//convert to TGV
int to_tgv(const char *infn, const char *outfn)
{
    static char buf[192*256+4*256];
    int r;

    cout << "[+] Create TGV file." << endl;
    r=create_tgv(outfn);
    if(!r)
    {
      cout << "Error: Cannot create output file.";
      return 0;
    }

    // process sound
    {
        pipe_t* hp = create_pipe();
        cout << "[+] Create pipe to ffmpeg (sound)." << endl;
        if(!hp)
        {
          cout << "Error: Cannot create named pipe.";
          return 0;
        }
        
        //create ffmpeg command
        ostringstream s;
        s << FFMPEG_BIN << " " <<
          "-i " << infn << " ";
        if(conf.bit_rate=="0")
        {
          cout << "[*] Copying audio stream as is! This might not work!" << endl;
          s << "-f mp3 -c:a copy ";
        }
        else
        {
          cout << "[*] Using conversion to MP3 with " << conf.bit_rate << " bitrate." << endl;
          s << "-f mp3 -b:a " << conf.bit_rate << " ";
        }
        s << "-vn " << conf.audio_parm << " " <<
          "-y " << hp->pipename;
          ;
    
        // run ffmpeg
        cout << "[+] Run ffmpeg: extract sound." << endl;
        proc_t* pp=run_proc((char*)(s.str().c_str()), !conf.same_con);
        if(!pp)
        {
          cout << "Error: Cannot create process.";
          return 0;
        }
    
        cout << "[+] Connecting to pipe." << endl;
        r=connect_pipe(hp);
        if(!r)
        {
          cout << "Error: Cannot connect to named pipe.";
          return 0;
        }

        //get sound data 
        cout << "[+] Receiving data..." << endl;
        int sz,snd_size=0;
        while( (sz=piperead(hp, buf, sizeof(buf))) )
        {
          write_tgv_snd(buf, sz);
          snd_size+=sz;
        }
        fin_tgv_snd();
        close_pipe(hp);
        wait_proc(pp);
        cout << "[+] Sound extracted, " << snd_size << " bytes." << endl;
        if(snd_size>2*1024*1024)
        {
          cout << "[!] WARNING: Sound exceeded 2M limit!" << endl;
        }
    }


    // process video
    {
        pipe_t* hp = create_pipe();
        cout << "[+] Create pipe to ffmpeg (video)." << endl;
        if(!hp)
        {
          cout << "Error: Cannot create named pipe.";
          return 0;
        }
        
        //create ffmpeg command
        ostringstream s;
        s.precision(17);
        s << FFMPEG_BIN << " " <<
          "-i " << infn << " " <<
          "-f image2pipe ";
        if(conf.filter=="")
        {
          s << "-vf scale=" << TGV_WIDTH << ":" << TGV_HEIGHT << ":force_original_aspect_ratio=decrease,pad=" << TGV_WIDTH << ":" << TGV_HEIGHT << ":(ow-iw)/2:(oh-ih)/2 ";
        }
        else
        {
          s << conf.filter;
        }
        cout << "[*] Using frame rate " << conf.frame_rate << " fps." << endl;
        s << "-r " << conf.frame_rate << " " <<
          "-an " <<
          "-pix_fmt pal8 " <<
          "-vcodec rawvideo " << conf.video_parm << " " <<
          "-y " << hp->pipename;
          ;
    
        // run ffmpeg
        cout << "[+] Run ffmpeg: extract video." << endl;
        proc_t* pp=run_proc((char*)(s.str().c_str()), !conf.same_con);
        if(!pp)
        {
          cout << "Error: Cannot create process.";
          return 0;
        }
    
        cout << "[+] Connecting to pipe." << endl;
        r=connect_pipe(hp);
        if(!r)
        {
          cout << "Error: Cannot connect to named pipe.";
          return 0;
        }

        //get video data
        cout << "[+] Receiving data..." << endl;
        unsigned char palbuf[256*2];
        int frames=0;
        while(piperead(hp, buf, sizeof(buf)))
        {
          unsigned char *palin=(unsigned char *) &buf[TGV_WIDTH*TGV_HEIGHT*1];
          unsigned char *palout=palbuf;
          int clr;
          for(int n=0;n<256;++n)
          {
            clr=(palin[2]/10)*1024 + (palin[1]/10)*32 + palin[0]/10;
            palin+=4;
            *palout++=clr&0xff;
            *palout++=clr>>8;
          }
          write_tgv_frame((char*)palbuf, 256*2);
          write_tgv_frame(buf, TGV_WIDTH*TGV_HEIGHT*1);
          frames++;
        }
        close_pipe(hp);
        wait_proc(pp);
        cout << "[+] Video extracted, " << frames << " frames." << endl;
    }

    //finalize TGV
    fin_tgv();
    cout << "[+] Conversion done!" << endl;

    return 1;
}


//convert from TGV
int from_tgv(const char *infn, const char *outfn)
{
    //open TGV file
    int r=open_tgv(infn);
    if(!r)
    {
      cout << "Error: Cannot open TGV file.";
      return 0;
    }

    pipe_t* hp = create_pipe();
    cout << "[+] Create pipe to ffmpeg." << endl;
    if(!hp)
    {
      cout << "Error: Cannot create named pipe.";
      return 0;
    }

    //extract mp3 to temp file
    int mp3sz;
    char *mp3buf = get_tgv_snd(mp3sz);

    FILE *t=fopen("temp.mp3","wb");
    fwrite(mp3buf,1,mp3sz,t);
    fclose(t);

    //create ffmpeg command
    ostringstream s;
    s.precision(17);
    if(conf.audio_parm=="")
      conf.audio_parm="-c:a copy";
    //if(conf.video_parm=="")
    //  conf.video_parm="-c:v libx264 -pix_fmt yuvj420p -preset:v veryfast";
    int frame_h = get_tgv_frame_h();

    s << FFMPEG_BIN << " " <<
      "-f mp3 " <<
      "-i temp.mp3 " <<
      "-f rawvideo " <<
      "-r " << conf.frame_rate << " " <<
      "-video_size " << TGV_WIDTH << "x" << frame_h << " " <<
      "-pix_fmt pal8 " <<
      "-vcodec rawvideo " <<
      "-y -i " << hp->pipename << " " <<
      conf.audio_parm << " " << conf.video_parm << " " <<
      outfn
      ;

    // run ffmpeg
    cout << "[+] Run ffmpeg: create video." << endl;
    proc_t* pp=run_proc((char*)(s.str().c_str()), !conf.same_con);
    if(!pp)
    {
      cout << "Error: Cannot create process.";
      return 0;
    }

    cout << "[+] Connecting to pipe." << endl;
    r=connect_pipe(hp);
    if(!r)
    {
      cout << "Error: Cannot connect to named pipe.";
      return 0;
    }

    //put video data
    cout << "[+] Sending data..." << endl;
    unsigned char palbuf[256*4];
    char *fr;
    int frames=0;
    while( (fr=get_tgv_frame()) )
    {
      unsigned char *palin=(unsigned char *) &fr[0];
      unsigned char *palout=palbuf;
      int clr;
      for(int n=0;n<256;++n)
      {
        clr=palin[0]+256*palin[1];
        palin+=2;
        *palout++=(clr&0x1f)*10;
        *palout++=((clr>>5)&0x1f)*10;
        *palout++=((clr>>10)&0x1f)*10;
        *palout++=0xff;
      }
    
      pipewrite(hp, &fr[256*2], TGV_WIDTH*frame_h*1);
      pipewrite(hp, (char*)palbuf, sizeof(palbuf));
      frames++;
    }
    close_pipe(hp);
    wait_proc(pp);
    cout << "[+] Video created, " << frames << " frames." << endl;

    //all done!
    close_tgv();
    cout << "[+] Conversion done!" << endl;

    return 1;
}







//YO!
int main(int argc, char **argv)
{
    //config defaults
    conf.frame_rate = TGV_FPS;
    conf.bit_rate = "128k";
    conf.same_con = false;
    conf.decode = false;
    conf.audio_parm = "";
    conf.video_parm = "";
    DEBUG_OUT.set_debug(0);
    cout.precision(17);

    // get options
    int c;
    opterr = 0;
    while ((c = getopt (argc, argv, "hf:r:b:dctA:V:")) != -1)
        switch (c)
        {
            case 'h':
                usage(argv[0]);
                return 0;
            case 'f':
                conf.filter = optarg;
                break;
            case 'r':
                conf.frame_rate = atof(optarg);
                break;
            case 'b':
                conf.bit_rate = optarg;
                break;
            case 'd':
                DEBUG_OUT.set_debug(1);
                break;
            case 'c':
                conf.same_con = true;
                break;
            case 't':
                conf.decode = true;
                break;
            case 'A':
                conf.audio_parm = optarg;
                break;
            case 'V':
                conf.video_parm = optarg;
                break;

            case '?':
                switch (optopt)
                {
                    case 'h':
                    case 'f':
                    case 'r':
                    case 'b':
                    case 'd':
                    case 'c':
                    case 't':
                    case 'A':
                    case 'V':
                        cerr << "Option -" << (char)optopt << " requires an argument." << endl;
                        break;
                    default:
                        usage(argv[0]);
                }
                return 1;
        }

    // get file names
    vector<string> infiles;
    for (int index = optind; index<argc; index++)
        infiles.push_back(argv[index]);

    if (infiles.size() < 2)
    {
        cerr << "Both input and output files must be specified. Try '-h' for help." << endl;
        return 1;
    }

    //decode or encode TGV
    if(conf.decode)
    {
      c=from_tgv(infiles[0].c_str(),infiles[1].c_str());
    }
    else
    {
      c=to_tgv(infiles[0].c_str(),infiles[1].c_str());
    }
    return c!=0;
}


